
(* ---------------------------------------------------------------------------------------------------------------------- *)
(*                                     Chargement et sauvegarde de fichier sans cache                                     *)
(* ---------------------------------------------------------------------------------------------------------------------- *)

Function CHARGEMENT_CLASSIQUE (
                              hWindow : hWnd;                 (* Handle de la fenêtre appelante *)
                              FichierSource : pChar;          (* Chemin complet du fichier à charger *)
                              var pTete, pQueue : pDonnees;   (* Tête et queue de la liste chaînée *)
                              var NbEnreg : Cardinal          (* Compteur d'enregistrements *)
                             ) : LongBool;                    (* True si pas d'erreur *)
(* Chargement de fichier dans une liste chaînée bidirectionnelle *)
Var AncienCurseur : hCursor;          (* Ancien curseur *)
   HandleFichier : tHandle;          (* Handle du fichier *)
   NbBytesLus : DWord;               (* Nombre de bytes lus dans le fichier lors de la dernière opération *)
   OkOuverture : LongBool;           (* Résultat de l'ouverture du fichier *)
   OkLecture : LongBool;             (* Résultat de la lecture du fichier *)
   OkAllocation : LongBool;          (* Résultat de l'allocation de mémoire dynamique *)
   FinFichier : LongBool;            (* Indique que la fin du fichier est atteinte *)
   OkFermeture : LongBool;           (* Résultat de la fermeture du fichier *)
   p : pDonnees;                     (* Elément à ajouter dans la liste chaînée *)
   Titre : Array [0..51] of Char;    (* Titre de la boîte à messages *)
   Texte : Array [0..299] of Char;   (* Texte de la boîte à messages - Buffer de lecture du cache *)
Begin
 AncienCurseur := SetCursor(LoadCursor(0,IDC_WAIT));
 HandleFichier := CreateFile(FichierSource,GENERIC_READ,FILE_SHARE_READ,Nil,OPEN_EXISTING,FILE_ATTRIBUTE_ARCHIVE,0);
 OkOuverture := (HandleFichier <> INVALID_HANDLE_VALUE);
 if OkOuverture
    then
      begin
        OkLecture := True;
        FinFichier := False;
        OkAllocation := True;
        repeat
          (* Allocation de l'élément suivant à insérer dans la boîte liste *)
          New(p);
          if p <> Nil
             then
               begin
                 Inc(NbEnreg);
                 FillChar(p^,SizeOf(p^),0);
                 (* Entier *)
                 if ReadFile(HandleFichier,p^.Entier,SizeOf(p^.Entier),NbBytesLus,Nil)
                    then
                      begin
                        if NbBytesLus < SizeOf(p^.Entier)
                           then
                             FinFichier := True;
                      end
                    else
                      OkLecture := False;
                 (* Chaîne *)
                 if OkLecture and not FinFichier
                    then
                      if ReadFile(HandleFichier,p^.Chaine,SizeOf(p^.Chaine),NbBytesLus,Nil)
                         then
                           begin
                             if NbBytesLus < SizeOf(p^.Chaine)
                                then
                                  FinFichier := True;
                           end
                         else
                           OkLecture := False;
                 (* Ajout dans la liste chaînée *)
                 if OkLecture and not FinFichier
                    then
                      begin
                        if pQueue = Nil
                           then
                             pTete := p
                           else
                             begin
                               pQueue^.pSuiv := p;
                               p^.pPrec := pQueue;
                             end;
                        pQueue := p;
                      end
                    else   (* Désallocation de l'élément *)
                      Dispose(p);
                 end
               else   (* Mémoire insuffisante *)
                 OkAllocation := False;
        until FinFichier or (not OkAllocation) or (not OkLecture);
      end;
 OkFermeture := CloseHandle(HandleFichier);
 SetCursor(AncienCurseur);
 (* Message d'erreur éventuel *)
 if not OkOuverture
    then
      StrCopy(Texte,'Impossible d''ouvrir le fichier')
    else
      if not OkLecture
         then
           StrCopy(Texte,'Erreur de lecture du fichier')
         else
           if not OkFermeture
              then
                StrCopy(Texte,'Erreur lors de la fermeture du fichier')
              else
                if not OkAllocation
                   then
                     StrCopy(Texte,'M'#233'moire insuffisante pour charger le fichier');
 if (not OkOuverture) or (not OkLecture) or (not OkFermeture) or (not OkAllocation)
    then
      begin
        StrCat(Texte,#10#13);
        StrCat(Texte,FichierSource);
        StrCopy(Titre,'Chargement de fichier');
        MessageBox(hWindow,Texte,Titre,mb_IconHand or mb_Ok);
      end;
 (* Résultat de la fonction *)
 CHARGEMENT_CLASSIQUE := OkOuverture and OkLecture and OkFermeture and OkAllocation;
End;
