
{ http://alcatiz.developpez.com/tutoriel/lecture-ecriture-fichier-cache/ }

Program DEMO_CACHE;

(* ---------------------------------------------------------------------------------------------------------------------- *)
(*                                                                                                                        *)
(*  Exemple de lecture/criture de fichier en utilisant un cache.                                                         *)
(*                                                                                                                        *)
(*  30-10-2011 par Alcatz pour Developpez.com.                                                                           *)
(*  Compilateur : Virtual Pascal 2.1 build 279.                                                                           *)
(*                                                                                                                        *)
(*  Pour la lecture, on alloue un cache ayant la mme taille que le fichier et on y charge tout le contenu du fichier en  *)
(*  une seule fois.                                                                                                       *)
(*  Pour l'criture, on calcule la taille du fichier  crer, on alloue un cache de cette taille, on y copie toutes les   *)
(*  donnes et on crit l'entiret du fichier en une seule fois.                                                         *)
(*                                                                                                                        *)
(* ---------------------------------------------------------------------------------------------------------------------- *)


//{$B-,D+,H-,I-,J+,P-,Q-,R-,S-,T-,V+,W-,X+,Z-}
//{&AlignCode+,AlignData+,AlignRec+,Asm-,Cdecl-,Comments-,Delphi+,Frame+,G5+}
//{&LocInfo+,Open32-,Optimise+,OrgName-,SmartLink+,Speed+,Use32+,ZD-}
//{$M 32768}
{$MODE DELPHI}

Uses Strings,   (* Chanes AZT *)
    //WinCRT,    (* Emulation de console *)
    Windows;   (* API Windows *)


Type (* Liste chane bidirectionnelle contenant les donnes  modifier *)

    pDonnees = ^tDonnees;

    tDonnees = Record
                 Entier : LongInt;
                 Chaine : {String}ShortString;
                 pPrec, pSuiv : pDonnees;
               end;


Const (* Tte et queue de la liste chane *)

     pTete : pDonnees = Nil;
     pQueue : pDonnees = Nil;


Var hWindow : HWnd;                 (* Handle de la fentre WinCRT *)
   p : pDonnees;                   (* Elment de la liste chane *)
   NbEnreg : Cardinal;             (* Nombre d'enregistrements *)
   HrDebut, HrFin : tSystemTime;   (* Heure de dbut et de fin pour le chronomtre *)


(* ---------------------------------------------------------------------------------------------------------------------- *)
(*                               Dsallocation de la liste chane utilise par le programme                              *)
(* ---------------------------------------------------------------------------------------------------------------------- *)

Procedure DESALLOCATION_LISTE (
                              var pTete, pQueue : pDonnees   (* Tte et queue de la liste chane *)
                             );
(* Dsallocation de la liste chane *)
Var p1, p2 : pDonnees;
Begin
 p1 := pTete;
 while p1 <> Nil do
   begin
     p2 := p1^.pSuiv;
     Dispose(p1);
     p1 := p2;
   end;
 pTete := Nil;
 pQueue := Nil;
End;


(* ---------------------------------------------------------------------------------------------------------------------- *)
(*                                                       Chronomtre                                                      *)
(* ---------------------------------------------------------------------------------------------------------------------- *)

Function TEMPS_ECOULE (
                      HrDebut, HrFin : tSystemTime   (* Mesure du dbut et de fin *)
                     ) : Cardinal;                   (* Diffrence en millisecondes *)
(* Retourne le temps coul entre deux mesures, en millisecondes.
  Par facilit, ne gre pas le cas o la mesure s'est effectue sur plus d'une seule journe. *)
Begin
 TEMPS_ECOULE := Cardinal(HrFin.wHour) * 3600000 +
                 Cardinal(HrFin.wMinute) * 60000 +
                 Cardinal(HrFin.wSecond) * 1000 +
                 Cardinal(HrFin.wMilliseconds) -
                 Cardinal(HrDebut.wHour) * 3600000 -
                 Cardinal(HrDebut.wMinute) * 60000 -
                 Cardinal(HrDebut.wSecond) * 1000 -
                 Cardinal(HrDebut.wMilliseconds);
End;


(* ---------------------------------------------------------------------------------------------------------------------- *)
(*                                   Chargement et sauvegarde de fichier via un cache                                     *)
(* ---------------------------------------------------------------------------------------------------------------------- *)

Function CACHE_LECTURE (
                       hWindow : hWnd;          (* Handle de la fentre appelante *)
                       Chemin : pChar;          (* Chemin complet du fichier  charger *)
                       var Cache : pChar;       (* Adresse du cache *)
                       var Taille : Cardinal    (* Taille du cache *)
                      ) : LongBool;             (* True si pas d'erreur *)
(* Alloue un cache mmoire et y charge l'entiret du fichier spcifi *)
Var HandleFichier : tHandle;          (* Handle du fichier *)
   NbBytesLus : DWord;               (* Nombre de bytes lus dans le fichier lors de la dernire opration *)
   OkOuverture : LongBool;           (* Rsultat de l'ouverture du fichier *)
   OkTaille : LongBool;              (* Lecture de la taille du fichier *)
   OkLecture : LongBool;             (* Rsultat de la lecture du fichier *)
   OkAllocation : LongBool;          (* Rsultat de l'allocation de mmoire dynamique *)
   OkFermeture : LongBool;           (* Rsultat de la fermeture du fichier *)
   Titre : Array [0..51] of Char;    (* Titre de la bote  messages *)
   Texte : Array [0..299] of Char;   (* Texte de la bote  messages *)
Begin
 Cache := Nil;
 OkTaille := True;
 OkOuverture := True;
 OkLecture := True;
 OkAllocation := True;
 OkFermeture := True;
 (* Ouverture du fichier *)
 HandleFichier := CreateFile(Chemin,GENERIC_READ,FILE_SHARE_READ,Nil,OPEN_EXISTING,FILE_ATTRIBUTE_ARCHIVE,0);
 OkOuverture := (HandleFichier <> INVALID_HANDLE_VALUE);
 if OkOuverture
    then
      begin
        (* Lecture de la taille du fichier *)
        Taille := GetFileSize(HandleFichier,Nil);
        OkTaille := (Taille <> $FFFFFFFF);
        if OkTaille
           then   (* Allocation du cache mmoire *)
             begin
               GetMem(Cache,Taille);
               OkAllocation := (Cache <> Nil);
               if OkAllocation
                  then   (* Copie de l'entiret du fichier dans le cache *)
                    begin
                      if ReadFile(HandleFichier,Cache^,Taille,NbBytesLus,Nil)
                         then
                           begin
                             if NbBytesLus < Taille
                                then
                                  OkLecture := False;
                           end
                         else
                           OkLecture := False;
                    end;
             end;
        (* Fermeture du fichier *)
        OkFermeture := CloseHandle(HandleFichier);
      end;
 (* Message d'erreur ventuel *)
 if not OkTaille
    then
      StrCopy(Texte,'Impossible de d'#233'terminer la taille du fichier')
    else
      if not OkOuverture
         then
           StrCopy(Texte,'Impossible d''ouvrir le fichier')
         else
           if not OkLecture
              then
                StrCopy(Texte,'Erreur de lecture du fichier')
              else
                if not OkFermeture
                   then
                     StrCopy(Texte,'Erreur lors de la fermeture du fichier')
                   else
                     if not OkAllocation
                        then
                          StrCopy(Texte,'M'#233'moire insuffisante pour cr'#233'er le cache de lecture du fichier');
 if (not OkTaille) or (not OkOuverture) or (not OkLecture) or (not OkFermeture) or (not OkAllocation)
    then
      begin
        StrCat(Texte,#10#13);
        StrCat(Texte,Chemin);
        StrCopy(Titre,'Chargement de fichier');
        MessageBox(hWindow,Texte,Titre,mb_IconHand or mb_Ok);
      end;
 (* Rsultat de la fonction *)
 CACHE_LECTURE := OkTaille and OkOuverture and OkLecture and OkFermeture and OkAllocation;
End;


Function CACHE_ECRITURE (
                        hWindow : hWnd;      (* Handle de la fentre appelante *)
                        Chemin : pChar;      (* Chemin complet du fichier  crire *)
                        var Cache : pChar;   (* Adresse du cache  sauvegarder *)
                        Taille : Cardinal    (* Taille du cache *)
                       ) : LongBool;         (* True si pas d'erreur *)
(* Ecrit le contenu du cache mmoire dans un fichier *)
Var HandleFichier : tHandle;          (* Handle du fichier *)
   NbBytesEcrits : DWord;            (* Nombre de bytes crits dans le fichier lors de la dernire opration *)
   OkCreation : LongBool;            (* Rsultat de la cration du fichier *)
   OkEcriture : LongBool;            (* Rsultat de l'criture du fichier *)
   OkFermeture : LongBool;           (* Rsultat de la fermeture du fichier *)
   Titre : Array [0..51] of Char;    (* Titre de la bote  messages *)
   Texte : Array [0..299] of Char;   (* Texte de la bote  messages *)
Begin
 OkCreation := True;
 OkEcriture := True;
 OkFermeture := True;
 HandleFichier := CreateFile(Chemin,GENERIC_WRITE,FILE_SHARE_WRITE,Nil,CREATE_ALWAYS,FILE_ATTRIBUTE_ARCHIVE,0);
 OkCreation := (HandleFichier <> INVALID_HANDLE_VALUE);
 if OkCreation
    then
      begin
        OkEcriture := WriteFile(HandleFichier,Cache^,Taille,NbBytesEcrits,Nil);
        OkFermeture := CloseHandle(HandleFichier);
      end;
 (* Message d'erreur ventuel *)
 if not OkCreation
    then
      StrCopy(Texte,'Erreur lors de la cr'#233'ation du fichier')
    else
      if not OkEcriture
         then
           StrCopy(Texte,'Erreur d'''#233'criture dans le fichier')
         else
           if not OkFermeture
              then
                begin
                  StrCopy(Texte,'Erreur lors de la fermeture du fichier');
                  StrCat(Texte,#10#13);
                  StrCat(Texte,Chemin);
                end;
 if (not OkCreation) or (not OkEcriture) or (not OkFermeture)
    then
      begin
        StrCat(Texte,#10#13);
        StrCat(Texte,Chemin);
        StrCopy(Titre,'Sauvegarde de fichier');
        MessageBox(hWindow,Texte,Titre,mb_IconHand or mb_Ok);
      end;
 (* Rsultat de la fonction *)
 CACHE_ECRITURE := OkCreation and OkEcriture and OkFermeture;
End;


Function CHARGEMENT (
                    hWindow : hWnd;                 (* Handle de la fentre appelante *)
                    FichierSource : pChar;          (* Chemin complet du fichier  charger *)
                    var pTete, pQueue : pDonnees;   (* Tte et queue de la liste chane *)
                    var NbEnreg : Cardinal          (* Compteur d'enregistrements *)
                   ) : LongBool;                    (* True si pas d'erreur *)
(* Chargement de fichier dans une liste chane bidirectionnelle *)
Var AncienCurseur : hCursor;          (* Ancien curseur *)
   Cache : pChar;                    (* Cache dans lequel le fichier est entirement charg *)
   Taille : Cardinal;                (* Taille du fichier *)
   Index : Cardinal;                 (* Index dans le parcours squentiel du cache *)
   Adresse : pChar;                  (* Adresse dans le cache depuis laquelle il faut transfrer les donnes *)
   OkCache : LongBool;               (* Indique si la mise en cache s'est bien droule *)
   OkAllocation : LongBool;          (* Rsultat de l'allocation de mmoire dynamique *)
   p : pDonnees;                     (* Elment  ajouter dans la liste chane *)
   Titre : Array [0..51] of Char;    (* Titre de la bote  messages *)
   Texte : Array [0..299] of Char;   (* Texte de la bote  messages *)
Begin
 AncienCurseur := SetCursor(LoadCursor(0,IDC_WAIT));
 (* Mise en cache du fichier *)
 OkAllocation := True;
 OkCache := CACHE_LECTURE(hWindow,FichierSource,pChar(Cache),Taille);
 if OkCache
    then
      begin
        NbEnreg := 0;
        Adresse := Cache;
        Index := 0;
        while (Index < Taille) and OkAllocation do
          begin
            (* Allocation de l'lment suivant  insrer dans la bote liste *)
            New(p);
            if p <> Nil
               then
                 begin
                   Inc(NbEnreg);
                   FillChar(p^,SizeOf(p^),0);
                   (* Entier *)
                   StrMove(pChar(@p^.Entier),Adresse,SizeOf(p^.Entier));
                   Inc(Index,SizeOf(p^.Entier));
                   Adresse := Cache + Index;
                   (* Chane *)
                   StrMove(pChar(@p^.Chaine),Adresse,SizeOf(p^.Chaine));
                   Inc(Index,SizeOf(p^.Chaine));
                   Adresse := Cache + Index;
                   (* Ajout dans la liste chane *)
                   if pQueue = Nil
                      then
                        pTete := p
                      else
                        begin
                          pQueue^.pSuiv := p;
                          p^.pPrec := pQueue;
                        end;
                   pQueue := p;
                 end
               else   (* Mmoire insuffisante *)
                 OkAllocation := False;
          end;
        (* Dsallocation du cache *)
        FreeMem(Cache,Taille);
     end;
 SetCursor(AncienCurseur);
 (* Message d'erreur ventuel (les erreurs lies au fichier sont affiches par Cache_Lecture) *)
 if not OkAllocation
    then
      begin
        StrCopy(Texte,'M'#233'moire insuffisante pour charger le fichier');
        StrCat(Texte,#10#13);
        StrCat(Texte,FichierSource);
        StrCopy(Titre,'Chargement de fichier');
        MessageBox(hWindow,Texte,Titre,mb_IconHand or mb_Ok);
      end;
 (* Rsultat de la fonction *)
 CHARGEMENT := OkCache and OkAllocation;
End;


Function TAILLE_CACHE_SAUVEGARDE (
                                 NbEnreg : Cardinal   (* Nombre d'enregistrements *)
                                ) : Cardinal;         (* Taille en octets *)
(* Calcule la taille du cache d'criture du fichier *)
Begin
 TAILLE_CACHE_SAUVEGARDE := NbEnreg * (SizeOf(Cardinal) + SizeOf({String}ShortString));
End;


Function SAUVEGARDE (
                    hWindow : hWnd;             (* Handle de la fentre appelante *)
                    FichierDest : pChar;        (* Chemin complet du fichier  sauvegarder *)
                    pTete, pQueue : pDonnees;   (* Tte et queue de la liste chane *)
                    NbEnreg : Cardinal          (* Nombre d'enregistrements *)
                   ) : LongBool;                (* True si pas d'erreur *)
(* Sauvegarde du contenu de la liste chane dans un fichier *)
Var AncienCurseur : hCursor;          (* Ancien curseur *)
   OkAllocation : LongBool;          (* Rsultat de l'allocation de mmoire dynamique *)
   OkEcriture : LongBool;            (* Rsultat de l'criture dans le fichier *)
   Cache : pChar;                    (* Cache dans lequel l'entiret du fichier est charge *)
   TailleCache : Cardinal;           (* Taille du cache *)
   Adresse : pChar;                  (* Adresse courante dans le cache *)
   OkCache : LongBool;               (* Indique si la mise en cache s'est bien droule *)
   Index : Cardinal;                 (* Index courant dans le cache *)
   //n : Cardinal;                     (* Compteur d'enregistrements *)
   p : pDonnees;                     (* Elment  sauvegarder *)
   Titre : Array [0..51] of Char;    (* Titre de la bote  messages *)
   Texte : Array [0..299] of Char;   (* Texte de la bote  messages - Buffer d'criture dans le fichier *)
Begin
 AncienCurseur := SetCursor(LoadCursor(0,IDC_WAIT));
 (* Cration du cache *)
 OkAllocation := True;
 TailleCache := TAILLE_CACHE_SAUVEGARDE(NbEnreg);
 GetMem(Cache,TailleCache);
 OkCache := (Cache <> Nil);
 if OkCache
    then   (* Mise en cache *)
      begin
        Adresse := Cache;
        Index := 0;
        p := pTete;
        while p <> Nil do
          begin
            (* Entier *)
            Adresse := Cache + Index;
            StrMove(Adresse,pChar(@p^.Entier),SizeOf(p^.Entier));
            Inc(Index,SizeOf(p^.Entier));
            (* Chane *)
            Adresse := Cache + Index;
            StrMove(Adresse,pChar(@p^.Chaine),SizeOf(p^.Chaine));
            Inc(Index,SizeOf(p^.Chaine));
            (* Passage  l'lment suivant *)
            p := p^.pSuiv;
          end;
        (* Ecriture du fichier *)
        OkEcriture := CACHE_ECRITURE(hWindow,FichierDest,Cache,TailleCache);
        (* Dsallocation du cache *)
        FreeMem(Cache,TailleCache);
      end
    else
      OkAllocation := False;
 SetCursor(AncienCurseur);
 (* Message d'erreur ventuel (les erreurs lies au fichier sont affiches par CACHE_ECRITURE) *)
 if not OkAllocation
    then
      begin
        StrCopy(Texte,'M'#233'moire insuffisante pour cr'#233'er le cache d'''#233'criture du fichier');
        Strcat(Texte,#10#13);
        StrCat(Texte,FichierDest);
        StrCopy(Titre,'Sauvegarde de fichier');
        MessageBox(hWindow,Texte,Titre,mb_IconHand or mb_Ok);
      end;
 (* Rsultat de la fonction *)
 SAUVEGARDE := OkCache and OkAllocation and OkEcriture;
End;


(* ---------------------------------------------------------------------------------------------------------------------- *)
(*                                                   Programme principal                                                  *)
(* ---------------------------------------------------------------------------------------------------------------------- *)

Begin
 //InitWinCRT;
 WriteLn('D'#233'monstration de lecture/'#233'criture de fichier via un cache.');
 WriteLn('30-10-2011 par Alcat'#238'z pour Developpez.com');
 Writeln;
 Write('Pressez Enter pour d'#233'marrer : ');
 ReadLn;
 WriteLn;
 (* Rcupration du handle de la fentre WinCRT *)
 hWindow := GetFocus;
 (* Chargement du contenu du fichier dans la liste chane *)
 WriteLn('Chargement du fichier... ');
 GetLocalTime(HrDebut);   (* Dmarrage du chrono *)
 if CHARGEMENT(hWindow,{'\\Machine_Distante\Repertoire\Demo.dat'}'demo.dat',pTete,pQueue,NbEnreg)
    then
      begin
        GetLocalTime(HrFin);   (* Arrt du chrono *)
        WriteLn('   OK - Dur'#233'e : ',TEMPS_ECOULE(HrDebut,HrFin),' millisecondes.');
        (* Modification des donnes *)
        p := pTete;
        while p <> Nil do
          begin
            p^.Chaine := p^.Chaine + ' a '#233't'#233' modifi'#233' !';
            p := p^.pSuiv;
          end;
        (* Sauvegarde des donnes modifies *)
        WriteLn('Sauvegarde du fichier... ');
        GetLocalTime(HrDebut);   (* Dmarrage du chrono *)
        if SAUVEGARDE(hWindow,{'\\Machine_Distante\Repertoire\DemoModif.dat'}'demomodif.dat',pTete,pQueue,NbEnreg)
           then
             begin
               GetLocalTime(HrFin);   (* Arrt du chrono *)
               WriteLn('   OK - Dur'#233'e : ',TEMPS_ECOULE(HrDebut,HrFin),' millisecondes.');
             end
           else
             WriteLn('   Erreur !');
        (* Dsallocation de la liste chane *)
        DESALLOCATION_LISTE(pTete,pQueue);
      end
    else
      WriteLn('   Erreur !');
 Write('Pressez Enter pour quitter : ');
 ReadLn;
 //DoneWinCRT;
End.
