/***************************************************************

  UciFilter.cpp

  Filtert die UCI-Kommunikation zwischen Arena und Engine

***************************************************************/

/* http://www.nnuss.de/Hermann/UciFilter.html */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <windows.h>
#include <ctype.h>
#include <time.h>

/* -------------------------------------------------------------------------- */

FILE *gLogFile1 = NULL;
FILE *gLogFile2 = NULL;

void LogLn(const char *line, const int extra)
{
  FILE *f;
  char buff[32];
  time_t now = time(NULL);
  struct tm *t = localtime(&now);
  strftime(buff, sizeof(buff) - 1, "%d/%m/%Y %H:%M:%S", t);
  f = (extra == 0) ? gLogFile1 : gLogFile2;
  fprintf(f, "%s %s\n", buff, line);
  fflush(f);
}

char gBuffer[256];

/* -------------------------------------------------------------------------- */

typedef long long Int_64;

typedef struct
{
  HANDLE hVonEngine;
  HANDLE hNachEngine;
  HANDLE hVonGui;
  HANDLE hNachGui;
  HANDLE hEngineNachFilter;
  HANDLE hEngine;
  HANDLE hThread;
  volatile BOOL bEngineRechnet;
  volatile BOOL bQuit;
  volatile Int_64 ZeitCurrmove;
  volatile Int_64 ZeitInfo;
  volatile long ValueCurrmove;
  volatile long ValueInfo;
  volatile long ValueQuit;
  volatile BOOL bFilterNonUci;
  volatile BOOL bFilterStop;
} FilterDaten;

typedef struct
{
  HANDLE hDatei;
  char *pPuffer;
  size_t Groesse;
  size_t AnzInhalt;
  size_t AnzVerarbeitet;
} Lesepuffer;

void FatalerFehler(const char *meldung, HANDLE hEngine);
Int_64 Jetzt(void);
void HandleDurchPipeErsetzen(DWORD EinOderAus, HANDLE *pMeinHandle, HANDLE *pEngineHandle, HANDLE *pOriginal);
HANDLE EngineStarten(int argc, char *argv[]);
void WriteOption(HANDLE h, const char *str);
char *StringLesen(Lesepuffer *pLP, HANDLE hEngine, const int iLog);
void GuiNachEngine(FilterDaten *pFilter);
DWORD WINAPI EngineNachGui(LPVOID pParam);
void EngineNachGuiThreadStarten(FilterDaten *pFilter);

void FatalerFehler(const char *meldung, HANDLE hEngine)
{
  DWORD LastError = GetLastError();
  fprintf(stderr, "info string UciFilter: FEHLER %ld; %s\n", (long)LastError, meldung);
  fflush(stderr);
  if (hEngine != NULL) TerminateThread(hEngine, 0);
  exit(0);
}

Int_64 Jetzt(void)
{
  FILETIME ft;
  LARGE_INTEGER li;
  GetSystemTimeAsFileTime(&ft);
  li.LowPart = ft.dwLowDateTime;
  li.HighPart = ft.dwHighDateTime;
  return li.QuadPart / 10000;
}

void HandleDurchPipeErsetzen(DWORD EinOderAus, HANDLE *pMeinHandle, HANDLE *pEngineHandle, HANDLE *pOriginal)
{
  SECURITY_ATTRIBUTES SecAttr;
  HANDLE hRein, hRaus, hDoppel, hWasBinIch;
  HANDLE *phMeins, *phEngineIhrs;
  
  sprintf(gBuffer, "line %03d %s", __LINE__, "-> HandleDurchPipeErsetzen()"); LogLn(gBuffer, 0);
  
  memset(&SecAttr, 0, sizeof(SecAttr));
  SecAttr.nLength = sizeof(SecAttr);
  SecAttr.bInheritHandle = TRUE;
  *pOriginal = GetStdHandle(EinOderAus);
  if (!CreatePipe(&hRein, &hRaus, &SecAttr, 0))
    FatalerFehler("CreatePipe geht nicht\n", NULL);
  if (EinOderAus == STD_OUTPUT_HANDLE) {
    phMeins = &hRein;
    phEngineIhrs = &hRaus;
  } else {
    phMeins = &hRaus;
    phEngineIhrs = &hRein;
  }
  if (!SetStdHandle(EinOderAus, *phEngineIhrs)) FatalerFehler("Ein- oder Ausgabe umbiegen geht nicht", NULL);
  hWasBinIch = GetCurrentProcess();
  if (!DuplicateHandle(hWasBinIch, *phMeins, hWasBinIch, &hDoppel, 0, FALSE, DUPLICATE_SAME_ACCESS))
    FatalerFehler("Doppeln fehlgeschlagen", NULL);
  CloseHandle(*phMeins);
  *pMeinHandle = hDoppel;
  *pEngineHandle = *phEngineIhrs;
  
  sprintf(gBuffer, "line %d %s", __LINE__, "<- HandleDurchPipeErsetzen()"); LogLn(gBuffer, 0);
}

HANDLE EngineStarten(int argc, char *argv[])
{
  PROCESS_INFORMATION ProcessInfo;
  STARTUPINFO StartInfo;
  char Kommandozeile[10000];
  char Dir[_MAX_DIR];
  char DirDrive[_MAX_PATH];
  int i;
  size_t laenge;
  
  sprintf(gBuffer, "line %d %s", __LINE__, "-> EngineStarten()"); LogLn(gBuffer, 0);
  
  memset(&ProcessInfo, 0, sizeof(ProcessInfo));
  memset(&StartInfo, 0, sizeof(StartInfo));
  StartInfo.cb = sizeof(StartInfo);
  Kommandozeile[0] = '\0';
  laenge = 0;
  for (i = 1; i < argc; i++)
  {
    if (laenge + strlen(argv[i]) + 1 >= sizeof(Kommandozeile)) FatalerFehler("Die Kommandozeile ist zu lang", NULL);
    if (i > 1) strcat(Kommandozeile, " ");
    strcat(Kommandozeile, argv[i]);
  }
  _splitpath(argv[1], DirDrive, Dir, NULL, NULL);
  strcat(DirDrive, Dir);
  
  sprintf(gBuffer, "line %d argv[1]='%s' Kommandozeile='%s' DirDrive='%s'", __LINE__, argv[1], Kommandozeile, DirDrive); LogLn(gBuffer, 0);
  
  if (!CreateProcess(
        argv[1],
        Kommandozeile,
        NULL,
        NULL,
        TRUE,
        0,
        NULL,
        DirDrive,
        &StartInfo,
        &ProcessInfo
      ))
    FatalerFehler("Engineprozess lsst sich nicht starten", NULL);
  
  sprintf(gBuffer, "line %d %s", __LINE__, "<- EngineStarten()"); LogLn(gBuffer, 0);
  
  return ProcessInfo.hProcess;
}

char *StringLesen(Lesepuffer *pLP, HANDLE hEngine, const int iLog)
{
  char *pAnfang, *pEnde;
  DWORD AnzGelesen;

  sprintf(gBuffer, "line %d -> StringLesen(%d, %d, %d)", __LINE__, pLP->Groesse, pLP->AnzInhalt, pLP->AnzVerarbeitet); LogLn(gBuffer, iLog);

  if (pLP->AnzInhalt > pLP->AnzVerarbeitet)
  {
    pAnfang = pLP->pPuffer + pLP->AnzVerarbeitet;
    pEnde = strchr(pLP->pPuffer + pLP->AnzVerarbeitet, '\n');
    
    sprintf(gBuffer, "line %d pLP->AnzInhalt=%d pLP->AnzVerarbeitet=%d", __LINE__, pLP->AnzInhalt, pLP->AnzVerarbeitet); LogLn(gBuffer, iLog);
    
    if (pEnde != NULL)
    {
      pLP->AnzVerarbeitet = pEnde - pLP->pPuffer + 1;
      assert(pLP->AnzInhalt >= pLP->AnzVerarbeitet);

      assert(isspace(*pEnde));
      while (isspace(*pEnde) && pEnde >= pAnfang)
        *pEnde-- = '\0';
      
      sprintf(gBuffer, "line %d %s", __LINE__, "<- StringLesen()");
      LogLn(gBuffer, iLog);
      
      return pAnfang;
    }
    else
    {
      if (pLP->AnzVerarbeitet > 0)
      {
        memmove(pLP->pPuffer, pAnfang, pLP->AnzInhalt - pLP->AnzVerarbeitet + 1);
        pLP->AnzInhalt -= pLP->AnzVerarbeitet;
        pLP->AnzVerarbeitet = 0;
      }
    }
  }
  else
  {
    pLP->AnzInhalt = 0;
    pLP->AnzVerarbeitet = 0;
  }

  do
  {
    if (!ReadFile(pLP->hDatei, pLP->pPuffer + pLP->AnzInhalt, pLP->Groesse - pLP->AnzInhalt, &AnzGelesen, NULL))
	  FatalerFehler("Lesefehler", hEngine);
    
    sprintf(gBuffer, "line %d AnzGelesen=%lu", __LINE__, AnzGelesen); LogLn(gBuffer, iLog);
    
    pLP->AnzInhalt += AnzGelesen;
    assert(pLP->AnzInhalt <= pLP->Groesse);
    pLP->pPuffer[pLP->AnzInhalt] = '\0';
    pAnfang = pLP->pPuffer + pLP->AnzVerarbeitet;
    pEnde = strchr(pAnfang, '\n');
    
    if (pEnde == NULL)
    {
      if (pLP->AnzInhalt == pLP->Groesse)
      {
        pLP->Groesse *= 2;
        pLP->pPuffer = realloc(pLP->pPuffer, pLP->Groesse + 1);

        if (pLP->pPuffer == NULL)
          FatalerFehler("Nicht gengend Speicher", hEngine);
      }
    }
  }
  while (pEnde == NULL);
  pLP->AnzVerarbeitet = pEnde - pLP->pPuffer + 1;
  assert(pLP->AnzInhalt >= pLP->AnzVerarbeitet);
  assert(isspace(*pEnde));
  while (isspace(*pEnde) && pEnde >= pAnfang) *pEnde-- = '\0';
  
  sprintf(gBuffer, "line %d %s", __LINE__, "<- StringLesen()"); LogLn(gBuffer, iLog);
  
  return pAnfang;
}

void GuiNachEngine(FilterDaten *pFilter)
{
  Lesepuffer lp;
  char Kommando[16];
  char *str;
  char *strName;
  char *strOption;
  char *strValue;
  size_t len;
  DWORD AnzGeschrieben;
  int i;
  
  sprintf(gBuffer, "line %d %s", __LINE__, "-> GuiNachEngine()"); LogLn(gBuffer, 0);
  
  lp.hDatei = pFilter->hVonGui;
  lp.Groesse = 10000;
  lp.pPuffer = malloc(lp.Groesse + 1); for (i = 0; i < sizeof(lp.pPuffer)/sizeof(lp.pPuffer[0]); i++) lp.pPuffer[i] = 0;
  lp.AnzInhalt = 0;
  lp.AnzVerarbeitet = 0;

  if (lp.pPuffer == NULL) FatalerFehler("Nicht gengend Speicherplatz", pFilter->hEngine);
  
  while (TRUE)
  {
    str = StringLesen(&lp, pFilter->hEngine, 0);
    len = strlen(str);
    sscanf(str, "%15s", Kommando);
    if (strcmp(Kommando, "setoption") == 0)
    {
      strName = strstr(str, " name ");
      strOption = strstr(str, "@UciFilter ");
      strValue = strstr(str, " value ");

      if (strName != NULL && strOption != NULL && strValue != NULL && strName < strOption && strOption < strValue)
      {
        if (strncmp(strOption - 9, " TimeQuit", 9) == 0)
        {
          sscanf(strValue + 7, "%ld", &pFilter->ValueQuit);
          continue;
        }
        else if (strncmp(strOption - 13, " TimeCurrmove", 13) == 0)
        {
          sscanf(strValue + 7, "%ld", &pFilter->ValueCurrmove);
          continue;
        }
        else if (strncmp(strOption - 9, " TimeInfo", 9) == 0)
        {
          sscanf(strValue + 7, "%ld", &pFilter->ValueInfo);
          continue;
        }
        else if (strncmp(strOption - 11, " FilterStop", 11) == 0)
        {
          strValue += 7;
          while (*strValue != '\0' && isspace(*strValue))
            strValue++;
          pFilter->bFilterStop = memcmp(strValue, "true", 4) == 0;
          continue;
        }
        else if (strncmp(strOption - 13, " FilterNonUci", 13) == 0)
        {
          strValue += 7;
          while (*strValue != '\0' && isspace(*strValue))
            strValue++;
          pFilter->bFilterNonUci = memcmp(strValue, "true", 4) == 0;
          continue;
        }
      }
    }
    else if (strcmp(Kommando, "go") == 0)
    {
      pFilter->bEngineRechnet = TRUE;
      pFilter->ZeitCurrmove = Jetzt();
      pFilter->ZeitInfo = pFilter->ZeitCurrmove;
      pFilter->ZeitCurrmove += pFilter->ValueCurrmove;
      pFilter->ZeitInfo += pFilter->ValueInfo;
    }
    else if (strcmp(Kommando, "stop") == 0)
    {
      if (pFilter->bFilterStop && !pFilter->bEngineRechnet)
        continue;
    }
    else if (strcmp(Kommando, "quit") == 0)
    {
      pFilter->bQuit = TRUE;

      WriteFile(pFilter->hEngineNachFilter, "\n", 1, &AnzGeschrieben, NULL);
      FlushFileBuffers(pFilter->hEngineNachFilter);
    }

    str[len] = '\n';

    if (!WriteFile(pFilter->hNachEngine, str, len + 1, &AnzGeschrieben, NULL) || AnzGeschrieben != len + 1) FatalerFehler("Schreibfehler", pFilter->hEngine);
    FlushFileBuffers(pFilter->hNachEngine);
    if (strcmp(Kommando, "quit") == 0)
    {
      if (WaitForSingleObject(pFilter->hEngine, pFilter->ValueQuit) != WAIT_OBJECT_0)
      {
        TerminateProcess(pFilter->hEngine, 0);
      }
      WaitForSingleObject(pFilter->hThread, 300);
      
      sprintf(gBuffer, "line %d %s", __LINE__, "<- GuiNachEngine()"); LogLn(gBuffer, 0);
	  
      return;
    }
  }
  
  sprintf(gBuffer, "line %d %s", __LINE__, "<- GuiNachEngine()"); LogLn(gBuffer, 0);
}

void WriteOption(HANDLE h, const char *str)
{
  DWORD AnzGeschrieben;
  
  sprintf(gBuffer, "line %d %s", __LINE__, "-> WriteOption()"); LogLn(gBuffer, 0);
  
  WriteFile(h, str, strlen(str), &AnzGeschrieben, NULL);
  
  sprintf(gBuffer, "line %d %s", __LINE__, "<- WriteOption()"); LogLn(gBuffer, 0);
}

DWORD WINAPI EngineNachGui(LPVOID pParam)
{
  FilterDaten *pFilter;
  Lesepuffer lp;
  char Kommando[16];
  char *str;
  size_t len;
  DWORD AnzGeschrieben;
  int i;
  
  sprintf(gBuffer, "line %d %s", __LINE__, "-> EngineNachGui()"); LogLn(gBuffer, 1);
  
  pFilter = (FilterDaten *)pParam;
  lp.hDatei = pFilter->hVonEngine;
  lp.Groesse = 10000;
  lp.pPuffer = malloc(lp.Groesse + 1); for (i = 0; i < sizeof(lp.pPuffer)/sizeof(lp.pPuffer[0]); i++) lp.pPuffer[i] = 0;
  lp.AnzInhalt = 0;
  lp.AnzVerarbeitet = 0;
  if (lp.pPuffer == NULL) FatalerFehler("Nicht gengend Speicherplatz", pFilter->hEngine);

  while (TRUE)
  {
    str = StringLesen(&lp, pFilter->hEngine, 1);
    
    len = strlen(str);
    if (pFilter->bQuit)
    {
      sprintf(gBuffer, "%d %s", __LINE__, "<- EngineNachGui()");
      LogLn(gBuffer, 1);
      return 0;
    }
    sscanf(str, "%15s", Kommando);

    if (strcmp(Kommando, "uciok") == 0)
    {
      WriteOption(pFilter->hNachGui, "option name TimeCurrmove@UciFilter type spin default 1500 min 0 max 5000\n");
      WriteOption(pFilter->hNachGui, "option name TimeInfo@UciFilter type spin default 300 min 0 max 5000\n");
      WriteOption(pFilter->hNachGui, "option name TimeQuit@UciFilter type spin default 1000 min 200 max 1800\n");
      WriteOption(pFilter->hNachGui, "option name RemoveNonUci@UciFilter type check default true\n");
      WriteOption(pFilter->hNachGui, "option name RemoveStop@UciFilter type check default true\n");
    }
    else if (strcmp(Kommando, "bestmove") == 0)
    {
      pFilter->bEngineRechnet = FALSE;
    }
    else if (strcmp(Kommando, "info") == 0)
    {
      if (strstr(str, " currmove ") != NULL)
      {
        if (Jetzt() < pFilter->ZeitCurrmove)
          continue;
      }
      else
      {
        if (Jetzt() < pFilter->ZeitInfo)
          continue;
      }
    }
    else if (pFilter->bFilterNonUci && strcmp(Kommando, "id") == 0 && strcmp(Kommando, "readyok") == 0 && strcmp(Kommando, "copyprotection") == 0 && strcmp(Kommando, "option") == 0)
    {
      continue;
    }

    str[len] = '\n';

    if (!WriteFile(pFilter->hNachGui, str, len + 1, &AnzGeschrieben, NULL) || AnzGeschrieben != len + 1)
      FatalerFehler("Schreibfehler", pFilter->hEngine);

    if (!pFilter->bEngineRechnet)
      FlushFileBuffers(pFilter->hNachGui);
  }
  
  sprintf(gBuffer, "line %d %s", __LINE__, "<- EngineNachGui()"); LogLn(gBuffer, 1);
}

void EngineNachGuiThreadStarten(FilterDaten *pFilter)
{
  ULONG ThreadID;
  
  sprintf(gBuffer, "line %d %s", __LINE__, "-> EngineNachGuiThreadStarten()"); LogLn(gBuffer, 0);
  
  pFilter->hThread = CreateThread(NULL, 256000, EngineNachGui, pFilter, 0, &ThreadID);
  
  sprintf(gBuffer, "line %d %s", __LINE__, "<- EngineNachGuiThreadStarten()"); LogLn(gBuffer, 0);
}

int main(int argc, char *argv[])
{
  FilterDaten Filter;
  HANDLE hEngineLesen;
  
  gLogFile1 = fopen("ucifilter.log", "w");
  gLogFile2 = fopen("ucifilter-extra.log", "w");
  
  sprintf(gBuffer, "line %d %s", __LINE__, "-> main()"); LogLn(gBuffer, 0);
  
  if (argc < 2)
    FatalerFehler("Keine Engine angegeben", NULL);
  
  memset(&Filter, 0, sizeof(Filter));
  Filter.ValueCurrmove = 1500;
  Filter.ValueInfo = 300;
  Filter.ValueQuit = 1000;
  Filter.bFilterStop = TRUE;
  Filter.bFilterNonUci = TRUE;
  HandleDurchPipeErsetzen(STD_OUTPUT_HANDLE, &Filter.hVonEngine, &Filter.hEngineNachFilter, &Filter.hNachGui);
  HandleDurchPipeErsetzen(STD_INPUT_HANDLE, &Filter.hNachEngine, &hEngineLesen, &Filter.hVonGui);
  Filter.hEngine = EngineStarten(argc, argv);
  if (!SetStdHandle(STD_INPUT_HANDLE, Filter.hVonGui) || !SetStdHandle(STD_OUTPUT_HANDLE, Filter.hNachGui)) FatalerFehler("Wiederherstellung von Standardein- oder -ausgabe fehlgeschlagen", Filter.hEngine);
  if (!CloseHandle(hEngineLesen)) FatalerFehler("Schlieen des lesenden Pipe-Endes bei Engine fehlgeschlagen", Filter.hEngine);
  EngineNachGuiThreadStarten(&Filter);
  GuiNachEngine(&Filter);
  CloseHandle(Filter.hEngineNachFilter);
  CloseHandle(Filter.hVonEngine);
  CloseHandle(Filter.hNachEngine);
  
  sprintf(gBuffer, "line %d %s", __LINE__, "<- main()"); LogLn(gBuffer, 0);
  
  fclose(gLogFile1);
  fclose(gLogFile2);
  
  return 0;
}
