
{ 
  
  Version en Pascal
  
  du programme UCI Filter de Volker Annuss
  
  http://www.nnuss.de/Hermann/UciFilter.html
  
}

program UciFilter;

{$ASSERTIONS ON}

uses
  SysUtils, Classes, Windows, Log;

function IsSpace(const c: ansichar): boolean;
begin
  result := (c in [#32, #9, #10, #11, #12, #13]);
end;

type
  FilterDaten = record
    hVonEngine, hNachEngine, hVonGui, hNachGui, hEngineNachFilter, hEngine, hThread: THandle;
    bEngineRechnet, bQuit: longbool;
    ZeitCurrmove, ZeitInfo: Int64;
    ValueCurrmove, ValueInfo, ValueQuit: integer;
    bFilterNonUci, bFilterStop: longbool;
  end;
  
  Lesepuffer = record
    hDatei: THandle;
    pPuffer: pansichar;
    Groesse, AnzInhalt, AnzVerarbeitet: longword;
  end;
  
  PLesepuffer = ^Lesepuffer;
  PFilterDaten = ^FilterDaten;

procedure FatalerFehler(const Meldung: pansichar; hEngine: THandle); forward;
function Jetzt(): Int64; forward;
procedure HandleDurchPipeErsetzen(EinOderAus: DWORD; pMeinHandle, pEngineHandle, pOriginal: PHandle); forward;
function EngineStarten(EnginePath: AnsiString): THandle; forward;
function StringLesen(pLP: PLesepuffer; hEngine: THandle; bExtraLog: boolean): pansichar; forward;
procedure GuiNachEngine(pFilter: PFilterDaten); forward;
procedure WriteOption(h: THandle; const str_: pansichar); forward;
function EngineNachGui(pParam: Pointer): DWORD; stdcall; forward;
procedure EngineNachGuiThreadStarten(pFilter: PFilterDaten); forward;

procedure FatalerFehler(const Meldung: pansichar; hEngine: THandle);
var
  LastError: DWORD;
begin
  LastError := GetLastError();
  WriteLn(ErrOutput, Format('info string UciFilter: FEHLER %d; %s', [LastError, Meldung]));
  Flush(ErrOutput);
  if hEngine <> 0 then TerminateThread(hEngine, 0);
  Halt(0);
end;

function Jetzt(): Int64;
var
  ft: FILETIME;
  li: LARGE_INTEGER;
begin
  GetSystemTimeAsFileTime(ft);
  
  li.LowPart := ft.dwLowDateTime;
  li.HighPart := ft.dwHighDateTime;
  
  result := li.QuadPart div 10000;
end;

procedure HandleDurchPipeErsetzen(EinOderAus: DWORD; pMeinHandle, pEngineHandle, pOriginal: PHandle);
var
  SecAttr: SECURITY_ATTRIBUTES;
  hRein, hRaus, hDoppel, hWasBinIch: THandle;
  phMeins, phEngineIhrs: PHandle;
begin
  TLog.Append(Format('line %0.3d %s', [StrToInt({$I %LINE%}), '-> HandleDurchPipeErsetzen()']));
  
  FillChar(SecAttr, SizeOf(SECURITY_ATTRIBUTES), #0);
  with SecAttr do begin
    nLength := SizeOf(SECURITY_ATTRIBUTES);
    bInheritHandle := TRUE;
    lpSecurityDescriptor := nil;
  end;
  pOriginal^ := GetStdHandle(EinOderAus);
  if not CreatePipe(hRein, hRaus, @SecAttr, 0) then
    FatalerFehler('CreatePipe geht nicht', 0);
  if EinOderAus = STD_OUTPUT_HANDLE then begin
    phMeins := @hRein;
    phEngineIhrs := @hRaus;
  end else begin
    phMeins := @hRaus;
    phEngineIhrs := @hRein;
  end;
  if not SetStdHandle(EinOderAus, phEngineIhrs^) then
    FatalerFehler('Ein- oder Ausgabe umbiegen geht nicht', 0);
  hWasBinIch := GetCurrentProcess();
  if not DuplicateHandle(hWasBinIch, phMeins^, hWasBinIch, @hDoppel, 0, FALSE, DUPLICATE_SAME_ACCESS) then FatalerFehler('Doppeln fehlgeschlagen', 0);
  CloseHandle(phMeins^);
  pMeinHandle^ := hDoppel;
  pEngineHandle^ := phEngineIhrs^;
  
  TLog.Append(Format('line %0.3d %s', [StrToInt({$I %LINE%}), '<- HandleDurchPipeErsetzen()']));
end;

function EngineStarten(EnginePath: AnsiString): {HANDLE}THandle;
var
  ProcessInfo: PROCESS_INFORMATION;
  StartInfo: _STARTUPINFOA;
  EngineDir: AnsiString;
begin
  TLog.Append(Format('120 %0.3d %s', [StrToInt({$I %LINE%}), '-> EngineStarten()']));
  
  FillChar(ProcessInfo, SizeOf(PROCESS_INFORMATION), #0);
  FillChar(StartInfo, SizeOf(_STARTUPINFOA), #0);
  StartInfo.cb := SizeOf(_STARTUPINFOA);
  
  EngineDir := ExtractFileDir(EnginePath);
  
  if not CreateProcess(
    nil,
    pansichar(EnginePath),
    nil,
    nil,
    TRUE,
    0,
    nil,
    pansichar(EngineDir),
    StartInfo,
    ProcessInfo
  ) then
    FatalerFehler('Engineprozess lässt sich nicht starten', 0);
  
  result := ProcessInfo.hProcess;
  
  TLog.Append(Format('line %0.3d %s', [StrToInt({$I %LINE%}), '<- EngineStarten()']));
end;

function StringLesen(pLP: PLesepuffer; hEngine: THandle; bExtraLog: boolean): pansichar;
var
  pAnfang, pEnde: pansichar;
  AnzGelesen: DWORD;
begin
  AnzGelesen := 0;

  TLog.Append(
    Format('line %0.3d -> StringLesen(%d, %d, %d)', [StrToInt({$I %LINE%}), pLP^.Groesse, pLP^.AnzInhalt, pLP^.AnzVerarbeitet]),
    bExtraLog
  );
  
  if pLP^.AnzInhalt > pLP^.AnzVerarbeitet then
  begin
    pAnfang := pLP^.pPuffer + pLP^.AnzVerarbeitet;
    pEnde := StrPos(pLP^.pPuffer + pLP^.AnzVerarbeitet, #10);
    
    if pEnde <> nil then
    begin
      pLP^.AnzVerarbeitet := pEnde - pLP^.pPuffer + 1;
      Assert(pLP^.AnzInhalt >= pLP^.AnzVerarbeitet);
      
      Assert(IsSpace(pEnde^));
      while (IsSpace(pEnde^) and (pEnde >= pAnfang)) do
      begin
        pEnde^ := #0;
        Dec(pEnde);
      end;
      TLog.Append(Format('line %0.3d %s', [StrToInt({$I %LINE%}), '<- StringLesen()']), bExtraLog);
      result := pAnfang;
      exit;
    end else
    begin
      if pLP^.AnzVerarbeitet > 0 then
      begin
        //Move(pAnfang^, pLP^.pPuffer^, pLP^.AnzInhalt - pLP^.AnzVerarbeitet + 1);
        StrMove(pLP^.pPuffer, pAnfang, pLP^.AnzInhalt - pLP^.AnzVerarbeitet + 1);
        
        pLP^.AnzInhalt := pLP^.AnzInhalt - pLP^.AnzVerarbeitet;
        pLP^.AnzVerarbeitet := 0;
      end;
    end;
  end else
  begin
    pLP^.AnzInhalt := 0;
    pLP^.AnzVerarbeitet := 0;
  end;
  
  repeat
    if pLP^.hDatei = INVALID_HANDLE_VALUE then
      Halt(2018);
    if not ReadFile(pLP^.hDatei, (pLP^.pPuffer + pLP^.AnzInhalt)^, pLP^.Groesse - pLP^.AnzInhalt, AnzGelesen, nil) then
      FatalerFehler('Lesefehler', hEngine);
    
    TLog.Append(Format('line %0.3d AnzGelesen=%d', [StrToInt({$I %LINE%}), AnzGelesen]), bExtraLog);
    
    pLP^.AnzInhalt := pLP^.AnzInhalt + AnzGelesen;
    Assert(pLP^.AnzInhalt <= pLP^.Groesse);
    pLP^.pPuffer[pLP^.AnzInhalt] := #0;
    pAnfang := pLP^.pPuffer + pLP^.AnzVerarbeitet;
    pEnde := StrPos(pAnfang, #10);
    
    if (pEnde = nil) and (pLP^.AnzInhalt = pLP^.Groesse) then
    begin
      pLP^.Groesse := pLP^.Groesse * 2;
      ReAllocMem(pLP^.pPuffer, pLP^.Groesse + 1);
      
      if pLP^.pPuffer = nil then
        FatalerFehler('Nicht genügend Speicher', hEngine);
    end;
  until pEnde <> nil;
  
  pLP^.AnzVerarbeitet := pEnde - pLP^.pPuffer + 1;
  Assert(pLP^.AnzInhalt >= pLP^.AnzVerarbeitet);
  
  Assert(IsSpace(pEnde^));
  while (IsSpace(pEnde^) and (pEnde >= pAnfang)) do
  begin
    pEnde^ := #0;
    Dec(pEnde);
  end;
  
  result := pAnfang;
  TLog.Append(Format('line %0.3d %s', [StrToInt({$I %LINE%}), '<- StringLesen()']), bExtraLog);
end;

procedure GuiNachEngine(pFilter: PFilterDaten);
var
  lp: Lesepuffer;
  Kommando: array[0..15] of ansichar;
  str_, strName, strOption, strValue: pansichar;
  len: longword;
  AnzGeschrieben: DWORD;
  newline: ansichar;
begin
  TLog.Append(Format('line %0.3d %s', [StrToInt({$I %LINE%}), '-> GuiNachEngine()']));
  
  lp.hDatei := pFilter^.hVonGui;
  lp.Groesse := 10000;
  lp.pPuffer := StrAlloc(lp.Groesse + 1);
  lp.AnzInhalt := 0;
  lp.AnzVerarbeitet := 0;
  if lp.pPuffer = nil then
    FatalerFehler('Nicht genügend Speicherplatz', pFilter^.hEngine);
  
  while TRUE do
  begin
    str_ := StringLesen(@lp, pFilter^.hEngine, FALSE);
    len := StrLen(str_);
    
    StrLCopy(Kommando, str_, SizeOf(Kommando) - 1);
    
    if StrComp(Kommando, 'setoption') = 0 then
    begin
      strName := StrPos(str_, ' name ');
      strOption := StrPos(str_, '@UciFilter ');
      strValue := StrPos(str_, ' value ');
      if (strName <> nil) and (strOption <> nil) and (strValue <> nil) and (strName < strOption) and (strOption < strValue) then
      begin
        if StrLComp(strOption - 9, ' TimeQuit', 9) = 0 then
        begin
          pFilter^.ValueQuit := StrToInt(StrPas(strValue + 7));
          continue;
        end
        else if StrLComp(strOption - 13, ' TimeCurrmove', 13) = 0 then
        begin
          pFilter^.ValueCurrmove := StrToInt(StrPas(strValue + 7));
          continue;
        end
        else if StrLComp(strOption - 9, ' TimeInfo', 9) = 0 then
        begin
          pFilter^.ValueInfo := StrToInt(StrPas(strValue + 7));
          continue;
        end
        else if StrLComp(strOption - 11, ' FilterStop', 11) = 0 then
        begin
          Inc(strValue, 7);
          while (strValue^ <> #0) and IsSpace(strValue^) do
            Inc(strValue);
          pFilter^.bFilterStop := CompareMem(strValue, pansichar('true'), 4);
          continue;
        end
        else if StrLComp(strOption - 13, ' FilterNonUci', 13) = 0 then
        begin
          Inc(strValue, 7);
          while (strValue^ <> #0) and IsSpace(strValue^) do
            Inc(strValue);
          pFilter^.bFilterNonUci := CompareMem(strValue, pansichar('true'), 4);
          continue;
        end;
      end;
    end else
    if StrComp(Kommando, 'go') = 0 then
    begin
      pFilter^.bEngineRechnet := TRUE;
      pFilter^.ZeitCurrmove := Jetzt();
      pFilter^.ZeitInfo := pFilter^.ZeitCurrmove;
      Inc(pFilter^.ZeitCurrmove, pFilter^.ValueCurrmove);
      Inc(pFilter^.ZeitInfo, pFilter^.ValueInfo);
    end else
    if StrComp(Kommando, 'stop') = 0 then
    begin
      if pFilter^.bFilterStop and not pFilter^.bEngineRechnet then
        continue;
    end else
    if StrComp(Kommando, 'quit') = 0 then
    begin
      pFilter^.bQuit := TRUE;
      newline := #10;
      WriteFile(pFilter^.hEngineNachFilter, {#10}newline, 1, AnzGeschrieben, nil);
      FlushFileBuffers(pFilter^.hEngineNachFilter);
    end;
    str_[len] := #10;
    if (not WriteFile(pFilter^.hNachEngine, str_, len + 1, AnzGeschrieben, nil))
    or (AnzGeschrieben <> len + 1) then
      FatalerFehler('Schreibfehler', pFilter^.hEngine );
    FlushFileBuffers( pFilter^.hNachEngine );
    if StrComp(Kommando, 'quit') = 0 then
    begin
      if WaitForSingleObject(pFilter^.hEngine, pFilter^.ValueQuit) <> WAIT_OBJECT_0 then
        TerminateProcess(pFilter^.hEngine, 0);
      WaitForSingleObject(pFilter^.hThread, 300);
      TLog.Append(Format('line %0.3d %s', [StrToInt({$I %LINE%}), '<- GuiNachEngine()']));
      exit;
    end;
  end;
  
  TLog.Append(Format('line %0.3d %s', [StrToInt({$I %LINE%}), '<- GuiNachEngine()']));
end;

procedure WriteOption(h: THandle; const str_: pansichar);
var
  AnzGeschrieben: DWORD;
begin
  TLog.Append(Format('line %0.3d %s', [StrToInt({$I %LINE%}), '-> WriteOption()']));
  WriteFile(h, str_, StrLen(str_), AnzGeschrieben, nil);
  TLog.Append(Format('line %0.3d %s', [StrToInt({$I %LINE%}), '<- WriteOption()']));
end;

function EngineNachGui(pParam: Pointer): DWORD; stdcall;
var
  pFilter: PFilterDaten;
  lp: Lesepuffer;
  Kommando: array[0..15] of ansichar;
  str_: pansichar;
  len: longword;
  AnzGeschrieben: DWORD;
begin
  TLog.Append(Format('line %0.3d %s', [StrToInt({$I %LINE%}), '-> EngineNachGui()']), TRUE);
  
  pFilter := PFilterDaten(pParam);
  
  lp.hDatei := pFilter^.hVonEngine;
  lp.Groesse := 10000;
  lp.pPuffer := StrAlloc(lp.Groesse + 1);
  FillChar(lp.pPuffer^, lp.Groesse * SizeOf(ansichar), #0);
  
  lp.AnzInhalt := 0;
  lp.AnzVerarbeitet := 0;
  if lp.pPuffer = nil then
    FatalerFehler('Nicht genügend Speicherplatz', pFilter^.hEngine);
  
  while TRUE do
  begin
    str_ := StringLesen(@lp, pFilter^.hEngine, TRUE);
    len := StrLen(str_);
    
    if pFilter^.bQuit then
    begin
      result := 0;
      TLog.Append(Format('line %0.3d %s', [StrToInt({$I %LINE%}), '<- EngineNachGui()']), TRUE);
      exit;
    end;
    
    StrLCopy(Kommando, str_, SizeOf(Kommando) - 1);
    
    if StrComp(Kommando, 'uciok') = 0 then
    begin
      WriteOption(pFilter^.hNachGui, 'option name TimeCurrmove@UciFilter type spin default 1500 min 0 max 5000'#10);
      WriteOption(pFilter^.hNachGui, 'option name TimeInfo@UciFilter type spin default 300 min 0 max 5000'#10);
      WriteOption(pFilter^.hNachGui, 'option name TimeQuit@UciFilter type spin default 1000 min 200 max 1800'#10);
      WriteOption(pFilter^.hNachGui, 'option name RemoveNonUci@UciFilter type check default true'#10);
      WriteOption(pFilter^.hNachGui, 'option name RemoveStop@UciFilter type check default true'#10);
    end else
    if StrComp(Kommando, 'bestmove') = 0 then
    begin
      pFilter^.bEngineRechnet := FALSE;
    end else
    if StrComp(Kommando, 'info') = 0 then
    begin
      if StrPos(str_, ' currmove ') <> nil then
      begin
        if Jetzt() < pFilter^.ZeitCurrmove then
          continue;
      end else
      begin
        if Jetzt() < pFilter^.ZeitInfo then
          continue;
      end;
    end else
    if pFilter^.bFilterNonUci
    and (StrComp(Kommando, 'id') = 0)
    and (StrComp(Kommando, 'readyok') = 0)
    and (StrComp(Kommando, 'copyprotection') = 0)
    and (StrComp(Kommando, 'option') = 0)
    then
      continue;
    
    str_[len] := #10;
    
    if (not WriteFile(pFilter^.hNachGui, str_, len + 1, AnzGeschrieben, nil)) or (AnzGeschrieben <> len + 1) then
      FatalerFehler('Schreibfehler', pFilter^.hEngine);

    if not pFilter^.bEngineRechnet then
      FlushFileBuffers(pFilter^.hNachGui);
  end;
  
  TLog.Append(Format('line %0.3d %s', [StrToInt({$I %LINE%}), '<- EngineNachGui()']), TRUE);
end;

procedure EngineNachGuiThreadStarten(pFilter: PFilterDaten);
var
  ThreadID: ULONG;
begin
  TLog.Append(Format('line %0.3d %s', [StrToInt({$I %LINE%}), '-> EngineNachGuiThreadStarten()']));
  
  pFilter^.hThread := CreateThread(nil, 256000, @EngineNachGui, pFilter, 0, ThreadID);
  
  TLog.Append(Format('line %0.3d %s', [StrToInt({$I %LINE%}), '<- EngineNachGuiThreadStarten()']));
end;

var
  Filter: FilterDaten;
  hEngineLesen: THandle;

begin
  TLog.Append(Format('line %0.3d %s', [StrToInt({$I %LINE%}), '-> Programme principal']));
  
  if ParamCount = 0 then
    FatalerFehler('Keine Engine angegeben', 0);
  
  FillChar(Filter, SizeOf(FilterDaten), #0);
  
  Filter.ValueCurrmove := 1500;
  Filter.ValueInfo :=  300;
  Filter.ValueQuit := 1000;
  Filter.bFilterStop := TRUE;
  Filter.bFilterNonUci := TRUE;
  
  HandleDurchPipeErsetzen(STD_OUTPUT_HANDLE, @Filter.hVonEngine, @Filter.hEngineNachFilter, @Filter.hNachGui);
  HandleDurchPipeErsetzen(STD_INPUT_HANDLE, @Filter.hNachEngine, @hEngineLesen, @Filter.hVonGui);
  
  Filter.hEngine := EngineStarten(ParamStr(1));
  
  if (not SetStdHandle(STD_INPUT_HANDLE, Filter.hVonGui))
  or (not SetStdHandle(STD_OUTPUT_HANDLE, Filter.hNachGui)) then
    FatalerFehler('Wiederherstellung von Standardein- oder -ausgabe fehlgeschlagen', Filter.hEngine);
  
  if not CloseHandle(hEngineLesen) then
    FatalerFehler('Schließen des lesenden Pipe-Endes bei Engine fehlgeschlagen', Filter.hEngine);
  
  EngineNachGuiThreadStarten(@Filter);
  GuiNachEngine(@Filter);
  
  CloseHandle(Filter.hEngineNachFilter);
  CloseHandle(Filter.hVonEngine);
  CloseHandle(Filter.hNachEngine);
  
  TLog.Append(Format('line %0.3d %s', [StrToInt({$I %LINE%}), '<- Programme principal']));
end.
